;==============================================================================

; file: "gsi.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(##include "../lib/header.scm")

;------------------------------------------------------------------------------

(define (##main)

  (##define-macro (gsi-or x) #t)

  (define (process-initialization-file)

    (define (try filename)
      (let ((init (##open-input-file filename)))
        (if init
          (begin
            (##close-port init)
            (##eval-top (##list '##include filename)
                        ##interaction-cte)
            #t)
          #f)))

    (or (try "gambc.scm")
        (let ((homedir (##path-expand "~")))
          (if homedir
            (try (##string-append homedir "gambc.scm"))))))

  (define (process-evaluate-options options)
    (let loop ((lst options))
      (if (##not (##null? lst))
        (let ((opt (##car lst)))
          (if (##string=? (##car opt) "e")
            (##eval-top (##read (##open-input-string (##cdr opt)))
                        ##interaction-cte))
          (loop (##cdr lst))))))

  (define (interpreter-batch-mode arguments)
    (let ((loaded-script? #f))
      (let loop ((lst arguments))
        (if (or loaded-script? (##null? lst))
          (##exit 0)
          (let ((filename (##car lst))
                (type-callback
                 (lambda (script?)
                   (if script?
                     (begin
                       (set! loaded-script? #t)
                       (set! ##processed-argv lst))))))
            (if (##string=? filename "-")
              (if (##port-isatty ##stdin)
                (##repl)
                (##load-source-from-port #f type-callback ##stdin))
              (##load filename type-callback #t #f))
            (loop (##cdr lst)))))))

  (define (compiler-batch-mode options arguments)

    (define (scm-file? file)
      (##member (##path-extension file) '(".scm" "")))

    (define (c-file? file)
      (##equal? (##path-extension file) ".c" #f))

    (let* ((opts (##map ##car options))
           (sym-opts (##map ##string->symbol opts)))

      (let loop1 ((lst arguments)
                  (nb-scm-files 0))
        (if (##not (##null? lst))

          (let ((arg (##car lst)))
            (loop1 (##cdr lst)
                   (if (scm-file? arg)
                     (##fixnum.+ nb-scm-files 1)
                     nb-scm-files)))

          (let* ((gen-c?
                  (##member "c" opts))
                 (gen-dynamic?
                  (##member "dynamic" opts))
                 (link?
                  (##not (or gen-c? gen-dynamic?)))
                 (output
                  (let ((x (##assoc "o" options)))
                    (cond ((##not x)
                           #f)
                          (gen-dynamic?
                           (##write-string
                            "*** WARNING -- Dynamic object file: \"o\" option ignored"
                            ##stderr)
                           (##newline ##stderr)
                           #f)
                          ((and (##not link?) (##fixnum.< 1 nb-scm-files))
                           (##write-string
                            "*** WARNING -- Multiple output files: \"o\" option ignored"
                            ##stderr)
                           (##newline ##stderr)
                           #f)
                          (else
                           (##cdr x))))))

            (let loop2 ((lst arguments)
                        (rev-roots '()))
              (if (##not (##null? lst))

                (let* ((file (##car lst))
                       (root (##path-strip-extension file)))
                  (cond ((scm-file? file)
                         (if (##fixnum.< 1 nb-scm-files)
                           (begin
                             (##display file ##stdout #f)
                             (##display ":" ##stdout #f)
                             (##newline ##stdout)))
                         (if gen-dynamic?
                           (if (##not (compile-file file sym-opts))
                             (##exit 1))
                           (let ((out (or (and (##not link?) output)
                                          (absent-obj))))
                             (if (##not (compile-file-to-c file sym-opts out))
                               (##exit 1))))
                       (loop2 (##cdr lst) (##cons root rev-roots)))
                      ((c-file? file)
                       (loop2 (##cdr lst) (##cons root rev-roots)))
                      (else
                       (##display file ##stdout #f)
                       (##display ": ignored (file not recognized)" ##stdout #f)
                       (##newline ##stdout)
                       (loop2 (##cdr lst) rev-roots))))

              (let* ((flat?
                      (##member "flat" opts))
                     (base
                      (let ((x (##assoc "l" options)))
                        (cond ((##not x)
                               #f)
                              ((or (##not link?) flat?)
                               (##write-string
                                "*** WARNING -- No incremental link: \"l\" option ignored"
                                ##stderr)
                               (##newline ##stderr)
                               #f)
                              (else
                               (##cdr x))))))

                (if link?
                  (if (##not (##null? rev-roots))
                    (let ((roots (##reverse rev-roots)))
                      (if flat?
                        (link-flat
                         roots
                         (or output (absent-obj)))
                        (link-incremental
                         roots
                         (or output (absent-obj))
                         (or base (absent-obj))))))
                  (if flat?
                    (begin
                      (##write-string
                       "*** WARNING -- \"c\" or \"dynamic\" option was specified: \"flat\" option ignored"
                       ##stderr)
                      (##newline ##stderr))))))))))))

  (define (interactive-mode)

    (##define-macro (banner)
      (string-append "Gambit Version " c#compiler-version))

    (##display (banner) ##stdout #f)
    (##newline ##stdout)
    (##newline ##stdout)
    (##repl))

  (define (pipe-mode)
    (let ((port ##stdin))
      (let loop ()
        (let ((expr (##read port)))
          (if (##eof-object? expr)
            (##exit 0)
            (let ((val (##eval-top expr
                                   ##interaction-cte)))
              (##write val ##stdout #t)
              (##newline ##stdout)
              (loop)))))))

  (define (##split-command-line options-with-arguments cont)

    (define (option? arg)
      (and (##fixnum.< 1 (##string-length arg))
           (##char=? (##string-ref arg 0) #\-)))

    (define (convert-option arg)
      (##substring arg 1 (##string-length arg)))

    (let loop1 ((arguments (##cdr (##argv)))
                (rev-options '()))
      (if (and (##pair? arguments) (option? (##car arguments)))

        (let ((opt (convert-option (##car arguments)))
              (rest (##cdr arguments)))
          (if (##member opt options-with-arguments)
            (if (##pair? rest)
              (loop1 (##cdr rest)
                     (##cons (##cons opt (##car rest)) rev-options))
              (begin
                (##write-string
                  "*** WARNING -- Missing argument for option \""
                  ##stderr)
                (##display opt ##stderr #f)
                (##write-string "\"" ##stderr)
                (##newline ##stderr)
                (loop1 rest rev-options)))
            (loop1 rest
                   (##cons (##cons opt #f) rev-options))))

        (cont (##reverse rev-options) arguments))))

  (define (split-options lst options)
    (if (##null? lst)
      '(())
      (let* ((x (##car lst))
             (opt (##car x)))
        (if (##member opt options)
          (let ((y (split-options (##cdr lst) options)))
            (##cons (##cons x (##car y)) (##cdr y)))
          (##cons '() lst)))))

  (let ((gambcdir (##path-expand "~~")))
    (if gambcdir
      (##load (##string-append gambcdir "gambc") (lambda (script?) #f) #f #f)))

  (##split-command-line
    '("e" "o" "l")
    (lambda (raw-options arguments)
      (let* ((options1 (split-options raw-options '("f" "i")))
             (main-options (##car options1))
             (skip-initialization-file? (##assoc "f" main-options))
             (force-interpreter? (##assoc "i" main-options))
             (other-options
              (if (gsi-or force-interpreter?)
                (split-options
                 (##cdr options1)
                 '("e"))
                (split-options
                 (##cdr options1)
                 '("c" "flat" "o" "l" "dynamic"
                   "verbose" "report" "expansion" "gvm"
                   "check" "force" "debug"))))
             (known-options
              (##car other-options))
             (unknown-options
              (##cdr other-options)))
        (if (##not (##null? unknown-options))
          (begin
            (##write-string "*** WARNING -- Unprocessed options: " ##stderr)
            (##write (##map ##car unknown-options) ##stderr #f)
            (##newline ##stderr)))
        (if (##not skip-initialization-file?)
          (process-initialization-file))
        (process-evaluate-options known-options)
        (cond ((##not (##null? arguments))
               (if (gsi-or force-interpreter?)
                 (interpreter-batch-mode arguments)
                 (compiler-batch-mode known-options arguments)))
              ((##port-isatty ##stdin)
               (interactive-mode))
              (else
               (pipe-mode)))))))

(##main)
